;;; $Id: matrix.scm 100 2005-12-16 20:35:49Z rlpm $

;;; A Matrix Library for MzScheme
;;; Author: Rory McGuire <rlpm at cs dot unm dot edu>

;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 2 of the
;;; License, or (at your option) any later version.
     
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.

;;; You can obtain a copy of the GPL at:
;;;   http://www.fsf.org/copyleft/gpl.html.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; This file requires functions defined in helpers.scm.

(define (matrix-transpose M)
  (apply map list M))

(define (matrix-sum a b)
  (map (lambda (x y) (map + x y)) a b))

(define (matrix-subtract a b)
  (matrix-sum a (map (lambda (x) (map - x)) b)))

(define (matrix-product a b)
  (map (lambda (c)
         (map (lambda (r)
                (apply + (map * c r)))
                       (matrix-transpose b)))
       a))

(define (matrix-multiply a s)
  (map (lambda (x) (map (lambda (y) (* y s)) x)) a))

(define (matrix-ident n)
  (map (lambda (y) (map (lambda (x) (if (= x y) 1 0)) (iota n))) (iota n)))

(define (matrix-isolate m n)
  (let* ((vec (list-ref m n))
         (base (list-ref vec n))
         (tick (map cons (map (lambda (x) (- x 1)) (iota (length m))) m)))
    (if (= 0 base)
        (error "ERROR using matrix-inverse: Singular matrix"))
    (map (lambda (x)
           (if (= n (car x))
               (map (lambda (z) (/ z base)) (cdr x))
               (let* ((row (cdr x))
                      (elem (list-ref row n))
                      (mul (/ elem base))
                      (sub (map (lambda (z) (* mul z)) vec))
                      (new (map - row sub)))
                 new)))
         tick)))

(define (matrix-inverse m)
  (if (not (= (length m) (length (car m))))
      (error "ERROR using matrix-inverse: Matrix must be square"))
  (map (iterate cdr (length m))
       (letrec ((loop
                 (lambda (ls m)
                   (if (null? ls)
                       m
                       (loop (cdr ls) (matrix-isolate m (- (car ls) 1)))))))
         (let ((l (length m)))
           (loop (iota l)
                 (map append m (matrix-ident l)))))))

(define (matrix-zeros x y)
  (map (lambda (j) (map (lambda (j) 0) j))
       (map (lambda (j) (iota y)) (iota x))))

(define (matrix-random x y n)
  (map (lambda (j) (map (lambda (j) (random n)) j))
       (map (lambda (j) (iota y)) (iota x))))

(define (list-elem-add1 ls i)
  (letrec ((loop
            (lambda (ls i acc)
              (if (= 0 i)
                  (append (reverse acc) (cons (+ 1 (car ls)) (cdr ls)))
                  (loop (cdr ls) (- i 1) (cons (car ls) acc))))))
    (loop ls i ())))

(define (matrix-elem-add1 m i j)
  (letrec ((loop
            (lambda (m i acc)
              (if (= 0 i)
                  (append (reverse acc)
                          (cons (list-elem-add1 (car m) j) (cdr m)))
                  (loop (cdr m) (- i 1) (cons (car m) acc))))))
    (loop m i ())))

(define (matrix-pseudoinverse m)
  (let ((mt (matrix-transpose m)))
    (matrix-product mt (matrix-inverse (matrix-product m mt)))))
